# Copyright (c) 2014, 2015 MIT Probabilistic Computing Project.
#
# This file is part of Venture.
#
# Venture is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# Venture is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with Venture.  If not, see <http://www.gnu.org/licenses/>.

"""Venture values.

The design currently lives in doc/type-system.md
"""
from collections import OrderedDict
from numbers import Number
import hashlib
import operator

import numpy as np # type: ignore

from venture.lite.exception import VentureTypeError
from venture.lite.exception import VentureValueError
from venture.lite.mlens import MLens
from venture.lite.typing import Tuple # Pylint doesn't understand type comments pylint: disable=unused-import
from venture.lite.typing import Union # Pylint doesn't understand type comments pylint: disable=unused-import
from venture.lite.typing import TYPE_CHECKING
from venture.lite.utils import ensure_python_float
import venture.lite.ensure_numpy as enp
import venture.value.dicts as v

if TYPE_CHECKING:
  from venture.lite.types import VentureType # Pylint doesn't understand type comments pylint: disable=unused-import

# TODO Define reasonable __str__ and/or __repr__ methods for all the
# values and all the types.

class VentureValue(object):
  __slots__ = ()
  """Base class of all Venture values."""
  ### "natural" representation and conversions
  def getNumber(self):
    # type: () -> float
    raise VentureTypeError("Cannot convert %s to number" % type(self))
  def getInteger(self):
    raise VentureTypeError("Cannot convert %s to integer" % type(self))
  def getAtom(self):
    raise VentureTypeError("Cannot convert %s to atom" % type(self))
  def getBool(self):
    # type: () -> bool
    raise VentureTypeError("Cannot convert %s to bool" % type(self))
  def getSymbol(self):
    # type: () -> str
    raise VentureTypeError("Cannot convert %s to symbol" % type(self))
  def getString(self):
    raise VentureTypeError("Cannot convert %s to string" % type(self))
  def getForeignBlob(self):
    raise VentureTypeError("Cannot convert %s to foreign blob" % type(self))
  def getPair(self):
    raise VentureTypeError("Cannot convert %s to pair" % type(self))
  # Convention for containers: if the elt_type argument is None,
  # return a Python container of VentureValue objects.  Otherwise,
  # apply the transformation given by the given type to the elements.
  def getArray(self, _elt_type=None):
    raise VentureTypeError("Cannot convert %s to array" % type(self))
  def getSimplex(self):
    raise VentureTypeError("Cannot convert %s to simplex" % type(self))
  def getDict(self):
    # type: () -> Dict[VentureValue, VentureValue]
    raise VentureTypeError("Cannot convert %s to dict" % type(self))
  def getMatrix(self):
    raise VentureTypeError("Cannot convert %s to matrix" % type(self))
  def getSymmetricMatrix(self):
    raise VentureTypeError(
      "Cannot convert %s to symmetric matrix" % type(self))
  def getSP(self):
    raise VentureTypeError("Cannot convert %s to sp" % type(self))
  def getEnvironment(self):
    raise VentureTypeError("Cannot convert %s to environment" % type(self))

  ### Stack representation
  def asStackDict(self, _trace=None):
    raise Exception("Cannot convert %s to a stack dictionary" % type(self))
  @staticmethod
  def fromStackDict(thing):
    # type: (object) -> VentureValue
    # XXX The precise type is a recursive union of lists,
    # VentureValues, and dicts with specific keys.  That seems
    # annoying to represent, so "object" will do for now.
    if isinstance(thing, list):
      # TODO Arrays or lists?
      return VentureArray([VentureValue.fromStackDict(val) for val in thing])
    elif isinstance(thing, VentureValue):
      return thing
    else:
      t = thing["type"]
      if t not in stackable_types:
        raise VentureTypeError('Invalid type "%s"' % t)
      return stackable_types[t].fromStackDict(thing)

  ### Comparison
  def compare(self, other):
    st = type(self)
    ot = type(other)
    if st == ot:
      return self.compareSameType(other)
    if st in venture_numeric_types and ot in venture_numeric_types:
      return stupidCompare(self.getNumber(), other.getNumber())
    if venture_types.index(st) < venture_types.index(ot) : return -1
    else: return 1 # We already checked for equality
  def compareSameType(self, _):
    raise Exception("Cannot compare %s" % type(self))
  def equal(self, other):
    if self is other:
      return True
    st = type(self)
    ot = type(other)
    if st == ot:
      return self.equalSameType(other)
    if st in venture_numeric_types and ot in venture_numeric_types:
      return self.getNumber() == other.getNumber()
    else:
      return False
  def equalSameType(self, other): return self.compareSameType(other) == 0

  ### Generic container methods
  def lookup(self, _):
    raise VentureTypeError("Cannot look things up in %s" % type(self))
  def lookup_grad(self, _index, _direction):
    raise VentureTypeError(
      "Cannot compute gradient of looking things up in %s" % type(self))
  def contains(self, _):
    raise VentureTypeError("Cannot look for things in %s" % type(self))
  def length(self):
    raise VentureTypeError("Cannot measure length of %s" % type(self))
  def take(self, _ct):
    raise VentureTypeError("Cannot take a prefix of %s" % type(self))

  def __eq__(self, other):
    if isinstance(other, VentureValue):
      return self.equal(other)
    else:
      return False

  def __ne__(self, other):
    return not self == other

  def __cmp__(self, other):
    if isinstance(other, VentureValue):
      return self.compare(other)
    else:
      return NotImplemented

  # Some Venture value types form a natural vector space over reals,
  # so overload addition, subtraction, and multiplication by scalars.
  # def __add__(self, other), and also __radd__, __neg__, __sub__,
  # __mul__, __rmul__, and dot

  # All Venture values have a (possibly trivial) cotangent space.  The
  # cotangent space is exposed through map_real, which constructs a
  # cotangent for a value given a function to which to feed the
  # value's current value at every dimension.
  def map_real(self, _f):
    return 0

  def expressionFor(self):
    # TODO Do we actually need to be able to override this anywhere?
    # Or does quote insertion work reliably now?
    return v.quote(self.asStackDict(None))

  def isValidCompoundForm(self):
    return False

  def asPythonList(self, _elt_type=None):
    # type: (VentureType) -> List[object]
    raise VentureTypeError("Cannot view %s as a Python list" % type(self))

def vv_dot_product(v1, v2):
  """Dot product of venture values taking into account that either may be
a symbolic zero.  TODO: Introduce the VentureZero value to uniformize
this."""
  if v1 is 0 or v2 is 0: return 0
  return v1.dot(v2)

class VentureNumber(VentureValue):
  __slots__ = ('number',)
  def __init__(self, number):
    # type: (Union[int, long, float]) -> None
    self.number = ensure_python_float(number)
  def __repr__(self):
    if hasattr(self, "number"):
      return "VentureNumber(%r)" % (self.number,)
    else:
      return "VentureNumber(uninitialized)"
  def __str__(self):
    if hasattr(self, "number"):
      return str(self.number)
    else:
      return "VentureNumber(uninitialized)"
  def getNumber(self):
    # type: () -> float
    return self.number
  def getInteger(self):
    return int(self.number)
  def getBool(self):
    return self.number

  def asStackDict(self, _trace=None):
    return v.number(self.number)
  @staticmethod
  def fromStackDict(thing):
    return VentureNumber(thing["value"])
  def compareSameType(self, other):
    return stupidCompare(self.number, other.number)
  def __hash__(self):
    return hash(self.number)
  def __add__(self, other):
    if other == 0:
      return self
    else:
      return VentureNumber(self.number + other.number)
  def __radd__(self, other):
    if other == 0:
      return self
    else:
      return VentureNumber(other.number + self.number)
  def __neg__(self):
    return VentureNumber(-self.number)
  def __sub__(self, other):
    if other == 0:
      return self
    else:
      return VentureNumber(self.number - other.number)
  def __mul__(self, other):
    # Assume other is a scalar
    return VentureNumber(self.number * ensure_python_float(other))
  def __rmul__(self, other):
    # Assume other is a scalar
    return VentureNumber(ensure_python_float(other) * self.number)
  def __abs__(self):
    return VentureNumber(abs(self.number))
  def __float__(self):
    # Returns a Python number. Implemented for testing. assert_not_equal
    # calls round(), which in turn calls float()
    return self.number
  def dot(self, other):
    assert isinstance(other, VentureNumber)
    return self.number * other.number
  def map_real(self, f):
    return VentureNumber(f(self.number))
  def real_lenses(self):
    class NumberLens(MLens):
      # Poor man's closure: pass the relevant thing from the lexical
      # scope directly.
      def __init__(self, vn):
        self.vn = vn
      def get(self):
        return self.vn.number
      def set(self, new):
        self.vn.number = ensure_python_float(new)
    return [NumberLens(self)]
  def expressionFor(self): return self.number

class VentureInteger(VentureValue):
  __slots__ = ('number',)
  def __init__(self,number):
    # type: (Union[int, long, float]) -> None
    assert isinstance(number, (int, long, float))
    self.number = int(number)
  def __repr__(self):
    if hasattr(self, "number"):
      return "VentureInteger(%s)" % self.number
    else:
      return "VentureInteger(uninitialized)"
  def __str__(self):
    if hasattr(self, "number"):
      return str(self.number)
    else:
      return "VentureInteger(uninitialized)"
  def getInteger(self):
    return self.number
  def getNumber(self):
    # type: () -> float
    return float(self.number)
  def getBool(self):
    return self.number
  def asStackDict(self, _trace=None):
    return v.integer(self.number)
  @staticmethod
  def fromStackDict(thing):
    return VentureInteger(thing["value"])
  def compareSameType(self, other):
    return stupidCompare(self.number, other.number)
  def __hash__(self):
    return hash(self.number)
  def __add__(self, other):
    if other == 0:
      return self
    else:
      assert isinstance(other, VentureInteger)
      return VentureInteger(self.number + other.number)
  def __radd__(self, other):
    if other == 0:
      return self
    else:
      assert isinstance(other, VentureInteger)
      return VentureInteger(other.number + self.number)
  def expressionFor(self):
    return self.number

def stupidCompare(thing, other):
  # number.__cmp__(other) works for ints but not floats.  Guido, WTF!?
  # strings don't have __cmp__ either? or lists?
  if thing < other: return -1
  elif thing > other: return 1
  else: return 0

def lexicographicBoxedCompare(thing, other):
  if len(thing) < len(other): return -1
  if len(thing) > len(other): return 1

  # else same length
  for x,y in zip(thing,other):
    if x.compare(y) != 0: return x.compare(y)

  return 0

def lexicographicUnboxedCompare(thing, other):
  if len(thing) < len(other): return -1
  if len(thing) > len(other): return 1

  # else same length
  for x,y in zip(thing,other):
    if stupidCompare(x,y) != 0: return stupidCompare(x,y)

  return 0

def lexicographicMatrixCompare(thing, other):
  ct_thing = reduce(operator.mul, np.shape(thing), 1)
  ct_other = reduce(operator.mul, np.shape(other), 1)
  if ct_thing == 0 and ct_other == 0: return 0
  shape_cmp = lexicographicUnboxedCompare(np.shape(thing), np.shape(other))
  if shape_cmp != 0: return shape_cmp

  # else shame shape
  if np.array_equal(thing, other): return 0
  # Hack for finding the first discrepant element, via
  # http://stackoverflow.com/questions/432112/is-there-a-numpy-function-to-return-the-first-index-of-something-in-an-array
  diffs = np.array(thing - other)
  diff_indexes = np.nonzero(diffs)
  first_diff = diffs[diff_indexes[0][0]][diff_indexes[0][0]]
  return stupidCompare(first_diff, 0)

def sequenceHash(seq):
  return reduce(lambda res, item: res * 37 + item, [hash(i) for i in seq], 1)

class VentureAtom(VentureValue):
  __slots__ = ('atom',)
  def __init__(self,atom):
    assert isinstance(atom, Number)
    self.atom = atom
  def __repr__(self):
    return "Atom(%s)" % self.atom
  def getAtom(self):
    return self.atom
  def asStackDict(self, _trace=None):
    return v.atom(self.atom)
  @staticmethod
  def fromStackDict(thing):
    return VentureAtom(thing["value"])
  def compareSameType(self, other):
    return stupidCompare(self.atom, other.atom)
  def __hash__(self):
    return hash(self.atom)
  def __add__(self, other):
    if other == 0 or other == VentureInteger(0):
      return self
    else:
      raise VentureTypeError("Cannot move %s by a non-zero displacement" % (self,))
  def expressionFor(self):
    return v.quote(self) # TODO Is this right?

class VentureBool(VentureValue):
  __slots__ = ('boolean',)
  def __init__(self,boolean):
    # type: (bool) -> None
    assert isinstance(boolean, bool) or isinstance(boolean, np.bool_)
    self.boolean = bool(boolean)
  def __repr__(self):
    return "Bool(%s)" % self.boolean
  def __str__(self):
    return str(self.boolean)
  def getBool(self):
    return self.boolean
  def asStackDict(self, _trace=None):
    return v.boolean(self.boolean)
  @staticmethod
  def fromStackDict(thing):
    return VentureBool(thing["value"])
  def compareSameType(self, other):
    return stupidCompare(self.boolean, other.boolean)
  def __hash__(self):
    return hash(self.boolean)
  def __add__(self, other):
    if other == 0 or other == VentureInteger(0):
      return self
    else:
      raise VentureTypeError("Cannot move %s by a non-zero displacement" % (self,))
  def expressionFor(self):
    return v.symbol("true") if self.boolean else v.symbol("false")

class VentureSymbol(VentureValue):
  __slots__ = ('symbol',)
  def __init__(self,symbol):
    # type: (str) -> None
    self.symbol = symbol
  def __repr__(self):
    return "Symbol(%s)" % self.symbol
  def __str__(self):
    return self.symbol
  def getSymbol(self):
    # type: () -> str
    return self.symbol
  def asStackDict(self, _trace=None):
    return v.symbol(self.symbol)
  @staticmethod
  def fromStackDict(thing):
    return VentureSymbol(thing["value"])
  def compareSameType(self, other):
    return stupidCompare(self.symbol, other.symbol)
  def __hash__(self):
    return hash(self.symbol)
  def __add__(self, other):
    if other == 0 or other == VentureInteger(0):
      return self
    else:
      raise VentureTypeError("Cannot move %s by a non-zero displacement" % (self,))
  def expressionFor(self):
    return v.quote(self.asStackDict(None))

# The difference between strings and symbols is that strings are
# self-evaluating
class VentureString(VentureValue):
  __slots__ = ('strng',)
  def __init__(self,strng):
    # type: (str) -> VentureString
    self.strng = strng
  def __repr__(self):
    return "Strng(%s)" % self.strng
  def __str__(self):
    return self.strng
  def getSymbol(self):
    # type: () -> str
    return self.strng
  def getString(self):
    return self.strng
  def asStackDict(self, _trace=None):
    return v.string(self.strng)
  @staticmethod
  def fromStackDict(thing):
    return VentureString(thing["value"])
  def compareSameType(self, other):
    return stupidCompare(self.strng, other.strng)
  def __hash__(self):
    return hash(self.strng)
  def __add__(self, other):
    if other == 0 or other == VentureInteger(0):
      return self
    else:
      raise VentureTypeError("Cannot move %s by a non-zero displacement" % (self,))
  def expressionFor(self):
    return v.quote(self.asStackDict(None))

class VentureForeignBlob(VentureValue):
  __slots__ = ('datum',)
  # TODO Think about the interaction of foreign blobs with trace
  # copying and serialization
  def __init__(self, datum):
    self.datum = datum
  def getForeignBlob(self):
    return self.datum
  def asStackDict(self, _trace=None):
    return v.blob(self.datum)
  @staticmethod
  def fromStackDict(thing):
    return VentureForeignBlob(thing["value"])
  def equalSameType(self, other):
    return self.datum == other.datum

class VentureNil(VentureValue):
  __slots__ = ()
  def __init__(self):
    # type: () -> None
    pass
  def __repr__(self):
    return "Nil"

  def getArray(self, elt_type=None):
    return VentureArray([]).getArray(elt_type)

  def asStackDict(self, _trace=None):
    return v.list([])
  @staticmethod
  def fromStackDict(_):
    return VentureNil()

  def compareSameType(self, _):
    return 0 # All Nils are equal
  def __hash__(self):
    return 0
  def __add__(self, other):
    if other == 0 or other == VentureInteger(0):
      return self
    else:
      raise VentureTypeError("Cannot move %s by a non-zero displacement" % (self,))

  def lookup(self, index):
    raise VentureValueError("Index out of bounds: too long by %s" % index)
  def contains(self, _obj):
    return False
  def size(self):
    return 0
  def take(self, ct):
    if ct < 1:
      return self
    else:
      raise VentureValueError("Index out of bounds: too long by %s" % ct)

  def expressionFor(self):
    return [v.symbol("list")]

  def isValidCompoundForm(self):
    return True
  def asPossiblyImproperList(self):
    return ([], None)
  def asPythonList(self, _elt_type=None):
    # type: (VentureType) -> List[object]
    return []

class VenturePair(VentureValue):
  __slots__ = ('first', 'rest')
  def __init__(self,(first,rest)):
    # type: (Tuple[Union[int, VentureValue], Union[int, VentureValue]]) -> None
    # TODO Maybe I need to be careful about tangent and cotangent
    # spaces after all.  A true pair needs to have a venture value
    # inside; a pair that represents the cotangent of something needs
    # to have cotangents; but cotangents permit symbolic zeros.
    if first != 0: assert isinstance(first, VentureValue)
    if rest != 0:  assert isinstance(rest, VentureValue)
    self.first = first
    self.rest = rest
  def __repr__(self):
    (list_, tail) = self.asPossiblyImproperList()
    if tail is None:
      return "VentureList(%r)" % list_
    else:
      return "VentureList(%r . %r)" % (list_, tail)
  def __str__(self):
    (list_, tail) = self.asPossiblyImproperList()
    base = ", ".join([str(l) for l in list_])
    if tail is None:
      return "[" + base + "]"
    else:
      return "[" + base + " . " + str(tail) + "]"

  def getPair(self):
    return (self.first,self.rest)
  def getArray(self, elt_type=None):
    (list_, tail) = self.asPossiblyImproperList()
    if tail is None:
      return VentureArray(list_).getArray(elt_type)
    else:
      try:
        end_array = tail.getArray(elt_type)
        front_array = VentureArray(list_).getArray(elt_type)
        return front_array + end_array
      except VentureTypeError:
        raise Exception("Cannot convert an improper list to array")

  def asStackDict(self, trace=None):
    (list_, tail) = self.asPossiblyImproperList()
    if tail is None:
      return v.list([val.asStackDict(trace) for val in list_])
    else:
      return v.improper_list([val.asStackDict(trace) for val in list_],
                             tail.asStackDict())
  @staticmethod
  def fromStackDict(thing):
    if thing["type"] == "improper_list":
      (list_, tail) = thing["value"]
      return pythonListToImproperVentureList(
        VentureValue.fromStackDict(tail),
        *[VentureValue.fromStackDict(val) for val in list_])
    else:
      return pythonListToVentureList([VentureValue.fromStackDict(val)
                                      for val in thing["value"]])

  def compareSameType(self, other):
    fstcmp = self.first.compare(other.first)
    if fstcmp != 0: return fstcmp
    else: return self.rest.compare(other.rest)
  def equalSameType(self, other):
    return self.first.equal(other.first) and self.rest.equal(other.rest)
  def __hash__(self):
    return hash(self.first) + 37*hash(self.rest)

  def lookup(self, index):
    try:
      ind = index.getNumber()
    except VentureTypeError:
      raise VentureValueError("Looking up non-number %r in a list" % index)
    if ind < 0: # Equivalent to floor for negative floats
      raise VentureValueError("Index out of bounds %s" % index)
    if ind < 1: # Equivalent to truncating for positive floats
      return self.first
    else:
      return self.rest.lookup(VentureNumber(ind - 1))
  def lookup_grad(self, index, direction):
    if direction == 0: return 0
    try:
      ind = index.getNumber()
    except VentureTypeError:
      raise VentureValueError("Looking up non-number %r in a list" % index)
    if ind < 1:
      return VenturePair((direction, 0))
    else:
      return VenturePair((0, self.rest.lookup_grad(VentureNumber(ind - 1),
                                                   direction)))
  def contains(self, obj): # Treat the list as a set
    if obj.equal(self.first):
      return True
    elif not isinstance(self.rest, VenturePair):
      # Notably, this means I am not checking whether the obj is the
      # last member of an improper list.
      return False
    else:
      return self.rest.contains(obj)
  def size(self): # Really, length
    return 1 + self.rest.size()
  def take(self, ind):
    if ind < 1:
      return VentureNil()
    else:
      return VenturePair((self.first, self.rest.take(ind-1)))

  def __add__(self, other):
    if other == 0 or other == VentureInteger(0):
      return self
    else:
      return VenturePair((self.first + other.first, self.rest + other.rest))
  def __radd__(self, other):
    if other == 0 or other == VentureInteger(0):
      return self
    else:
      return VenturePair((other.first + self.first, other.rest + self.rest))
  def __neg__(self):
    return VenturePair((-self.first, -self.rest))
  def __sub__(self, other):
    if other == 0 or other == VentureInteger(0):
      return self
    else:
      return VenturePair((self.first - other.first, self.rest - other.rest))
  def __mul__(self, other):
    # Assume other is a scalar
    assert isinstance(other, Number)
    return VenturePair((self.first * other, self.rest * other))
  def __rmul__(self, other):
    # Assume other is a scalar
    assert isinstance(other, Number)
    return VenturePair((other * self.first, other * self.rest))
  def dot(self, other):
    return vv_dot_product(self.first, other.first) + \
      vv_dot_product(self.rest, other.rest)
  def map_real(self, f):
    return VenturePair((self.first.map_real(f), self.rest.map_real(f)))

  def expressionFor(self):
    return [v.symbol("pair"), self.first.expressionFor(),
            self.rest.expressionFor()]

  def isValidCompoundForm(self):
    return self.rest.isValidCompoundForm()
  def asPythonList(self, elt_type=None):
    # type: (VentureType) -> List[object]
    if elt_type is not None:
      return [elt_type.asPython(self.first)] + self.rest.asPythonList(elt_type)
    else:
      return [self.first] + self.rest.asPythonList()
  def asPossiblyImproperList(self):
    front = [self.first]
    rest = self.rest
    while isinstance(rest, VenturePair):
      front.append(rest.first)
      rest = rest.rest
    if isinstance(rest, VentureNil):
      return (front, None)
    else:
      return (front, rest)

def pythonListToVentureList(l):
  # type: (List) -> Union[VenturePair, VentureNil]
  return reduce(lambda t, h: VenturePair((h, t)), reversed(l), VentureNil())

def pythonListToImproperVentureList(tail, *l):
  return reduce(lambda t, h: VenturePair((h, t)), reversed(l), tail)

class VentureArray(VentureValue):
  """Venture arrays are heterogeneous, with O(1) access and O(n) copy."""
  __slots__ = ('array',)
  def __init__(self, array):
    self.array = array
  def __repr__(self):
    return "VentureArray(%s)" % (self.array,)
  def __str__(self):
    return "[" + ", ".join([str(l) for l in self.array]) + "]"
  def getArray(self, elt_type=None):
    if elt_type is None: # No conversion
      return self.array
    else:
      return [elt_type.asPython(val) for val in self.array]

  def compareSameType(self, other):
    return lexicographicBoxedCompare(self.array, other.array)
  def __hash__(self): return sequenceHash(self.array)

  def asStackDict(self, trace=None):
    return v.array([val.asStackDict(trace) for val in self.array])
  @staticmethod
  def fromStackDict(thing):
    return VentureArray([VentureValue.fromStackDict(val)
                         for val in thing["value"]])

  def lookup(self, index):
    try:
      ind = index.getNumber()
    except VentureTypeError:
      raise VentureValueError("Looking up non-number %r in an array" % index)
    if 0 <= int(ind) and int(ind) < len(self.array):
      return self.array[int(ind)]
    else:
      raise VentureValueError("Index out of bounds: %s" % index)
  def lookup_grad(self, index, direction):
    try:
      ind = index.getNumber()
    except VentureTypeError:
      raise VentureValueError("Looking up non-number %r in an array" % index)
    if 0 <= int(ind) and int(ind) < len(self.array):
      ans = VentureArray([direction if i == int(ind) else 0
                          for (i,_) in enumerate(self.array)])
      return ans
    else:
      raise VentureValueError("Index out of bounds: %s" % index)
  def contains(self, obj):
    # Not Python's `in` because I need to use custom equality
    # TODO I am going to have to overload the equality for dicts
    # anyway, so might as well eventually use `in` here.
    return any(obj.equal(li) for li in self.array)
  def size(self):
    return len(self.array)
  def take(self, ind):
    return VentureArray(self.array[0:ind])

  def __add__(self, other):
    if other == 0 or other == VentureInteger(0):
      return self
    else:
      return VentureArray([x + y for (x,y) in zip(self.array, other.array)])
  def __radd__(self, other):
    if other == 0 or other == VentureInteger(0):
      return self
    else:
      return VentureArray([y + x for (x,y) in zip(self.array, other.array)])
  def __neg__(self):
    return VentureArray([-x for x in self.array])
  def __sub__(self, other):
    if other == 0 or other == VentureInteger(0):
      return self
    else:
      return VentureArray([x - y for (x,y) in zip(self.array, other.array)])
  def __mul__(self, other):
    # Assume other is a scalar
    assert isinstance(other, Number)
    return VentureArray([x * other for x in self.array])
  def __rmul__(self, other):
    # Assume other is a scalar
    assert isinstance(other, Number)
    return VentureArray([other * x  for x in self.array])
  def dot(self, other):
    return sum([vv_dot_product(x, y)
                for (x,y) in zip(self.array, other.array)])
  def map_real(self, f):
    return VentureArray([x.map_real(f) for x in self.array])

  def expressionFor(self):
    return [v.symbol("array")] + [val.expressionFor() for val in self.array]

  def isValidCompoundForm(self): return True
  def asPythonList(self, elt_type=None):
    # type: (VentureType) -> List[object]
    # I don't think there is any way to tell mypy that the output is a
    # list of VentureValue if the input is None
    return self.getArray(elt_type)

class VentureArrayUnboxed(VentureValue):
  """Venture arrays of unboxed objects are homogeneous, with O(1) access
and O(n) copy."""
  __slots__ = ('array', 'elt_type')
  def __init__(self, array, elt_type):
    from venture.lite.types import VentureType
    if not isinstance(elt_type, VentureType):
      raise VentureTypeError("%s of %s is not a VentureType" %
                             (elt_type, type(elt_type)))
    self.elt_type = elt_type
    self.array = enp.ensure_numpy_if_possible(elt_type, array)
  def __repr__(self):
    return "VentureArrayUnboxed(%s)" % self.array

  # This method produces results that are at least Python-boxed
  def getArray(self, elt_type=None):
    if elt_type is None: # Convert to VentureValue
      return [self.elt_type.asVentureValue(val) for val in self.array]
    else:
      # TODO What I really need here is the function that converts
      # from the Python representation of self.elt_type to the Python
      # representation of elt_type, with an optimization if that
      # function is the identity.
      if elt_type == self.elt_type:
        return self.array
      else:
        return [elt_type.asPython(self.elt_type.asVentureValue(val))
                for val in self.array]

  def compareSameType(self, other):
    return lexicographicUnboxedCompare(self.array, other.array)
  def __hash__(self): return sequenceHash(self.array)

  def asStackDict(self,_trace=None):
    return v.array_unboxed(self.array, self.elt_type)
  @staticmethod
  def fromStackDict(thing):
    if thing["type"] == "vector":
      # TODO HACK: treat Puma's vectors as unboxed arrays of numbers,
      # because Puma can't make a NumberType to stick in the stack
      # dict.
      from venture.lite.types import NumberType
      return VentureArrayUnboxed(thing["value"], NumberType())
    return VentureArrayUnboxed(thing["value"], thing["subtype"])

  def lookup(self, index):
    try:
      ind = index.getNumber()
    except VentureTypeError:
      raise VentureValueError("Looking up non-number %r in an array" % index)
    if 0 <= int(ind) and int(ind) < len(self.array):
      return self.elt_type.asVentureValue(self.array[int(ind)])
    else:
      raise VentureValueError("Index out of bounds: %s" % index)
  def lookup_grad(self, index, direction):
    # TODO Really this should be an unboxed array of the gradient
    # types of the elements, with generic zeroes.
    try:
      ind = index.getNumber()
    except VentureTypeError:
      raise VentureValueError("Looking up non-number %r in an array" % index)
    if 0 <= int(ind) and int(ind) < len(self.array):
      ans = VentureArrayUnboxed(
        [self.elt_type.asPython(direction) if i == int(ind) else 0
         for (i,_) in enumerate(self.array)],
        self.elt_type)
      return ans
    else:
      raise VentureValueError("Index out of bounds: %s" % index)
  def contains(self, obj):
    return any(obj.equal(self.elt_type.asVentureValue(li))
               for li in self.array)
  def size(self):
    return len(self.array)
  def take(self, ind):
    return VentureArrayUnboxed(self.array[0:ind], self.elt_type)

  def __add__(self, other):
    if other == 0 or other == VentureInteger(0):
      return self
    else:
      return VentureArrayUnboxed(
        *enp.map2(operator.add, self.array, self.elt_type,
                  other.array, other.elt_type))
  def __radd__(self, other):
    if other == 0 or other == VentureInteger(0):
      return self
    else:
      return VentureArrayUnboxed(
        *enp.map2(operator.add, other.array, other.elt_type,
                  self.array, self.elt_type))
  def __neg__(self):
    return VentureArrayUnboxed(
      enp.map(operator.neg, self.array, self.elt_type), self.elt_type)
  def __sub__(self, other):
    if other == 0 or other == VentureInteger(0):
      return self
    else:
      return VentureArrayUnboxed(
        *enp.map2(operator.sub, self.array, self.elt_type,
                  other.array, other.elt_type))
  def __mul__(self, other):
    if isinstance(other, Number):
      # If other is a scalar, broadcast
      return VentureArrayUnboxed(
        enp.map(lambda x: x * other, self.array, self.elt_type), self.elt_type)
    else:
      # Otherwise, multiply pointwise
      return VentureArrayUnboxed(
        *enp.map2(operator.mul, self.array, self.elt_type,
                  other.array, other.elt_type))
  def __rmul__(self, other):
    # If other is a scalar, broadcast
    if isinstance(other, Number):
      return VentureArrayUnboxed(
        enp.map(lambda x: other * x, self.array, self.elt_type), self.elt_type)
    else:
      # Otherwise, multiply pointwise
      return VentureArrayUnboxed(
        *enp.map2(operator.mul, other.array, other.elt_type,
                  self.array, self.elt_type))
  def dot(self, other):
    return enp.dot(self.array, self.elt_type, other.array, other.elt_type)
  def map_real(self, f):
    # TODO Ascertain whether self.elt_type represents a real number or
    # not so I can do
    # return VentureArrayUnboxed(
    #   enp.map(f, self.array, self.elt_type), self.elt_type)
    # but the general case is
    new_type = self.elt_type.gradient_type()
    new_data = [new_type.asPython(self.elt_type.asVentureValue(val).map_real(f))
                for val in self.array]
    return VentureArrayUnboxed(new_data, new_type)

  def isValidCompoundForm(self): return False
  def asPythonList(self, elt_type=None):
    # type: (VentureType) -> List[object]
    return self.getArray(elt_type)

class VentureSimplex(VentureValue):
  """Simplexes are homogeneous unboxed arrays of probabilities.  They
are also supposed to sum to 1, but we are not checking that.

  """
  __slots__ = ('simplex',)
  def __init__(self,simplex):
    self.simplex = simplex
  def __repr__(self):
    return "VentureSimplex(%s)" % self.simplex
  def getArray(self, elt_type=None):
    from venture.lite.types import ProbabilityType
    # TODO Abstract similarities between this and the getArray method
    # of ArrayUnboxed
    if elt_type is None: # Convert to VentureValue
      return [ProbabilityType().asVentureValue(val) for val in self.simplex]
    else:
      if elt_type.__class__ == ProbabilityType:
        return self.simplex
      else:
        return [elt_type.asPython(ProbabilityType().asVentureValue(val))
                for val in self.simplex]
  def getSimplex(self): return self.simplex

  def compareSameType(self, other):
    # The Python ordering is lexicographic first, then by length, but
    # I think we want lower-d simplexes to compare less than higher-d
    # ones regardless of the point.
    return lexicographicUnboxedCompare(self.simplex, other.simplex)
  def __hash__(self):
    return sequenceHash(self.simplex)

  def asStackDict(self, _trace=None):
    # TODO As what type to reflect simplex points to the stack?
    return v.simplex(self.simplex)
  @staticmethod
  def fromStackDict(thing): return VentureSimplex(thing["value"])

  def lookup(self, index):
    from venture.lite.types import ProbabilityType
    return ProbabilityType().asVentureValue(self.simplex[index.getNumber()])
  def contains(self, obj):
    # Homogeneous; TODO make it return False instead of exploding for
    # non-numeric objects.
    return obj.getNumber() in self.simplex
  def size(self): return len(self.simplex)
  def take(self, ind):
    from venture.lite.types import ProbabilityType
    return VentureArrayUnboxed(self.simplex[0:ind], ProbabilityType())

  def expressionFor(self):
    return [v.symbol("simplex")] + self.simplex
  def map_real(self, f):
    # The cotangent space actually has the constraint that the
    # cotangents sum to 0
    from venture.lite.types import NumberType
    return VentureArrayUnboxed([f(p) for p in self.simplex], NumberType())

  def asPythonList(self, elt_type=None):
    from venture.lite.types import ProbabilityType
    if elt_type is None:
      return [ProbabilityType().asVentureValue(val) for val in self.simplex]
    else:
      if elt_type == ProbabilityType():
        return self.simplex
      else:
        return [elt_type.asPython(ProbabilityType().asVentureValue(val))
                for val in self.simplex]

class VentureDict(VentureValue):
  __slots__ = ('dict',)
  def __init__(self,d):
    # type: (Dict[VentureValue, VentureValue]) -> None
    assert isinstance(d, OrderedDict)
    self.dict = d

  def getDict(self):
    # type: () -> Dict[VentureValue, VentureValue]
    return self.dict

  def asStackDict(self, _trace=None):
    return v.dict([(key.asStackDict(_trace), val.asStackDict(_trace))
                   for (key, val) in self.dict.items()])
  @staticmethod
  def fromStackDict(thing):
    f = VentureValue.fromStackDict
    return VentureDict(OrderedDict((f(key), f(val)) for (key, val) in thing["value"]))

  def equalSameType(self, other):
    return len(set(self.dict.iteritems()) ^ set(other.dict.iteritems())) == 0

  def lookup(self, key):
    return self.dict[key]
  def contains(self, key):
    return key in self.dict
  def size(self): return len(self.dict)

  def expressionFor(self):
    (keys, vals) = zip(*self.dict.iteritems())
    return [v.symbol("dict"),
            [v.symbol("list")] + [key.expressionFor() for key in keys],
            [v.symbol("list")] + [val.expressionFor() for val in vals]]

# 2D array of numbers backed by a numpy array object
class VentureMatrix(VentureValue):
  __slots__ = ('matrix',)
  def __init__(self,matrix):
    self.matrix = np.asarray(matrix)
  def __repr__(self):
    return "VentureMatrix(%s)" % self.matrix

  def getMatrix(self):
    return self.matrix
  def getSymmetricMatrix(self):
    if matrixIsSymmetric(self.matrix):
      return self.matrix
    else:
      raise VentureTypeError("Matrix is not symmetric %s" % self.matrix)

  def compareSameType(self, other):
    return lexicographicMatrixCompare(self.matrix, other.matrix)
  def __hash__(self):
    # From http://stackoverflow.com/questions/5386694/fast-way-to-hash-numpy-objects-for-caching
    b = self.matrix.view(np.uint8)
    return hash(hashlib.sha1(b).hexdigest())

  def asStackDict(self, _trace=None):
    return v.matrix(self.matrix)
  @staticmethod
  def fromStackDict(thing):
    return VentureMatrix(thing["value"])

  def lookup(self, index):
    try:
      (vind1, vind2) = index.getPair()
      ind1 = vind1.getNumber()
      ind2 = vind2.getNumber()
    except VentureTypeError:
      raise VentureValueError("Looking up non-pair-of-numbers %r in a matrix" % index)
    from venture.lite.types import NumberType
    return NumberType().asVentureValue(self.matrix[int(ind1), int(ind2)])

  def __add__(self, other):
    if other == 0 or other == VentureInteger(0):
      return self
    else:
      return VentureMatrix(self.matrix + other.matrix)
  def __radd__(self, other):
    if other == 0 or other == VentureInteger(0):
      return self
    else:
      return VentureMatrix(other.matrix + self.matrix)
  def __neg__(self):
    return VentureMatrix(-self.matrix)
  def __sub__(self, other):
    if other == 0 or other == VentureInteger(0):
      return self
    else:
      return VentureMatrix(self.matrix - other.matrix)
  def __mul__(self, other):
    # Assume other is a scalar
    assert isinstance(other, Number)
    return VentureMatrix(self.matrix * other)
  def __rmul__(self, other):
    # Assume other is a scalar
    assert isinstance(other, Number)
    return VentureMatrix(other * self.matrix)
  def dot(self, other):
    assert isinstance(other, VentureMatrix)
    return np.sum(np.multiply(self.matrix, other.matrix))
  def map_real(self, f):
    # Do I seriously have to special-case this?
    if self.matrix.size == 0: return self
    return VentureMatrix(np.vectorize(f)(self.matrix))

  def expressionFor(self):
    return [v.symbol("matrix"),
            [v.symbol("list")] + [[v.symbol("list")] + [val for val in row]
                                  for row in self.matrix]]

class VentureSymmetricMatrix(VentureMatrix):
  __slots__ = ('matrix',)
  def __init__(self, matrix):
    self.matrix = np.asarray(matrix)
    assert matrixIsSymmetric(self.matrix)
  def __repr__(self):
    return "VentureSymmetricMatrix(%s)" % self.matrix

  def asStackDict(self, _trace=None):
    return v.symmetric_matrix(self.matrix)
  @staticmethod
  def fromStackDict(thing):
    return VentureSymmetricMatrix(thing["value"])

  def __add__(self, other):
    if other == 0 or other == VentureInteger(0):
      return self
    if isinstance(other, VentureSymmetricMatrix):
      return VentureSymmetricMatrix(self.matrix + other.matrix)
    else:
      return VentureMatrix(self.matrix + other.matrix)
  def __radd__(self, other):
    if other == 0 or other == VentureInteger(0):
      return self
    if isinstance(other, VentureSymmetricMatrix):
      return VentureSymmetricMatrix(other.matrix + self.matrix)
    else:
      return VentureMatrix(other.matrix + self.matrix)
  def __neg__(self):
    return VentureSymmetricMatrix(-self.matrix)
  def __sub__(self, other):
    if other == 0 or other == VentureInteger(0):
      return self
    if isinstance(other, VentureSymmetricMatrix):
      return VentureSymmetricMatrix(self.matrix - other.matrix)
    else:
      return VentureMatrix(self.matrix - other.matrix)
  def __mul__(self, other):
    # Assume other is a scalar
    assert isinstance(other, Number)
    return VentureSymmetricMatrix(self.matrix * other)
  def __rmul__(self, other):
    # Assume other is a scalar
    assert isinstance(other, Number)
    return VentureSymmetricMatrix(other * self.matrix)
  def map_real(self, f):
    # Do I seriously have to special-case this?
    if self.matrix.size == 0: return self
    candidate = np.vectorize(f)(self.matrix)
    return VentureSymmetricMatrix( (candidate + candidate.T) / 2. )

  def expressionFor(self):
    return v.quote(self.asStackDict(None))

def matrixIsSymmetric(matrix):
  return np.allclose(matrix.transpose(), matrix)

class SPRef(VentureValue):
  __slots__ = ('makerNode',)
  def __init__(self,makerNode):
    self.makerNode = makerNode
  def asStackDict(self, trace=None):
    assert trace is not None
    return trace.madeSPRecordAt(self.makerNode).asStackDict()

  @staticmethod
  def fromStackDict(thing): return thing["value"]
  # SPRefs are intentionally not comparable until we decide otherwise

# Actually, Environments and SPs have cotangent spaces too, in their
# own funny way, but we're not doing that now.

## SPs and Environments as well
## Not Requests, because we do not reflect on them

venture_types = [
  VentureNumber, VentureInteger, VentureAtom, VentureBool,
  VentureSymbol, VentureString, VentureForeignBlob, VentureNil, VenturePair,
  VentureArray, VentureArrayUnboxed, VentureSimplex, VentureDict,
  VentureMatrix, VentureSymmetricMatrix, SPRef]
  # Break load order dependency by not adding SPs and Environments yet

venture_numeric_types = [VentureNumber, VentureInteger]

stackable_types = OrderedDict([
  ("number", VentureNumber),
  ("real", VentureNumber),
  ("integer", VentureInteger),
  ("probability", VentureNumber), # TODO Eliminate this stack dict type?
  ("atom", VentureAtom),
  ("boolean", VentureBool),
  ("symbol", VentureSymbol),
  ("string", VentureString),
  ("blob", VentureForeignBlob),
  ("list", VenturePair),
  ("improper_list", VenturePair),
  ("vector", VentureArrayUnboxed),
  ("array", VentureArray),
  ("array_unboxed", VentureArrayUnboxed),
  ("simplex", VentureSimplex),
  ("dict", VentureDict),
  ("matrix", VentureMatrix),
  ("symmetric_matrix", VentureSymmetricMatrix),
  ("SP", SPRef), # As opposed to VentureSPRecord?
])

def registerVentureType(t, name = None):
  if t in venture_types: pass
  else:
    venture_types.append(t)
    if name is not None:
      stackable_types[name] = t
